#include <sys/time.h>
#include <unistd.h>

#include <glib.h>

#include <utillib.h>

#include "downmand-marshal.h"
#include "transfer-thread.h"

GType           transfer_thread_get_type       ();
static void     transfer_thread_class_init     (TransferThreadClass *klass);
static void     transfer_thread_init           (TransferThread *thread);

static gpointer parent_class = NULL;

void    transfer_thread_open_target_uri     (GnomeVFSAsyncHandle *handle,
                                             GnomeVFSResult result,
                                             gpointer data);
void    transfer_thread_open_source_uri     (GnomeVFSAsyncHandle *handle,
                                             GnomeVFSResult result,
                                             gpointer data);
void    transfer_thread_source_read         (GnomeVFSAsyncHandle *handle,
                                             GnomeVFSResult result,
                                             gpointer buffer,
                                             GnomeVFSFileSize bytes_requested,
                                             GnomeVFSFileSize bytes_read,
                                             gpointer data);
void    transfer_thread_target_write        (GnomeVFSAsyncHandle *handle,
                                             GnomeVFSResult result,
                                             gconstpointer buffer,
                                             GnomeVFSFileSize bytes_requested,
                                             GnomeVFSFileSize bytes_written,
                                             gpointer data);
gboolean    transfer_thread_restart_transfering
                                            (gpointer data);


enum {
    TARGET_CREATED_SIGNAL,
    TRANSFERING_SIGNAL,
    STOP_SIGNAL,
    DONE_SIGNAL,
    ERROR_SIGNAL,
    LAST_SIGNAL
};

static gint transfer_thread_signals[LAST_SIGNAL] = { 0 };

GType
transfer_thread_get_type ()
{
    static GType transfer_thread_type = 0;

    if (!transfer_thread_type)
    {
        static const GTypeInfo transfer_thread_info =
        {
            sizeof (TransferThreadClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) transfer_thread_class_init,
            NULL,           /* class_finalize */
            NULL,           /* class_data */
            sizeof (TransferThread),
            0,              /* n_preallocs */
            (GInstanceInitFunc) transfer_thread_init
        };

        transfer_thread_type = g_type_register_static (G_TYPE_OBJECT,
                                                       "TransferThread",
                                                       &transfer_thread_info,
                                                       0);
    }

    return transfer_thread_type;
}

static void
transfer_thread_class_init (TransferThreadClass *klass)
{
    parent_class = g_type_class_peek_parent (klass);

    transfer_thread_signals[TARGET_CREATED_SIGNAL] = g_signal_new (
            "target_created_signal",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (TransferThreadClass, target_created_signal),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    transfer_thread_signals[TRANSFERING_SIGNAL] = g_signal_new (
            "transfering_signal",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (TransferThreadClass, transfering_signal),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    transfer_thread_signals[STOP_SIGNAL] = g_signal_new (
            "stop_signal",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (TransferThreadClass, stop_signal),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    transfer_thread_signals[DONE_SIGNAL] = g_signal_new (
            "done_signal",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (TransferThreadClass, done_signal),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    transfer_thread_signals[ERROR_SIGNAL] = g_signal_new (
            "error_signal",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (TransferThreadClass, error_signal),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);
}

static void
transfer_thread_init (TransferThread *thread)
{
    thread->source_handle = NULL;
    thread->target_handle = NULL;

    thread->bytes_start = 0;
    thread->bytes_stop = 0;
    thread->bytes_done = 0;

    thread->speedlimit = 0;
    thread->oldtime.tv_sec = 0;

    thread->stop = FALSE;
    thread->create_target = FALSE;
}

TransferThread *
transfer_thread_new()
{
    return TRANSFER_THREAD (g_object_new (transfer_thread_get_type (), NULL));
}

void
transfer_thread_start (TransferThread *transfer_thread)
{
    transfer_thread->buffer = g_new (char, 4100);

    if (transfer_thread->create_target)
    {
        gnome_vfs_async_create_uri (&transfer_thread->target_handle,
                                    transfer_thread->target_uri,
                                    GNOME_VFS_OPEN_WRITE
//FIXME: seek support in method
                                    | GNOME_VFS_OPEN_RANDOM,
                                    FALSE,
//FIXME: use source perms, or umask...
                                    0666,
                                    GNOME_VFS_PRIORITY_DEFAULT,
                                    transfer_thread_open_target_uri,
                                    transfer_thread);
    }
    else
    {
        gnome_vfs_async_open_uri (&transfer_thread->target_handle,
                                  transfer_thread->target_uri,
                                  GNOME_VFS_OPEN_WRITE
//FIXME: seek support in method
                                  | GNOME_VFS_OPEN_RANDOM,
                                  GNOME_VFS_PRIORITY_DEFAULT,
                                  transfer_thread_open_target_uri,
                                  transfer_thread);
    }        
}

void
transfer_thread_open_target_uri (GnomeVFSAsyncHandle *handle,
                                 GnomeVFSResult result,
                                 gpointer data)
{
    TransferThread *transfer_thread = TRANSFER_THREAD (data);

    if (result != GNOME_VFS_OK)
    {
        transfer_thread->status = ERROR_OPENING_TARGET;
        g_signal_emit (G_OBJECT (transfer_thread),
                       transfer_thread_signals[ERROR_SIGNAL],
                       0);
        return;
    }

    g_signal_emit (G_OBJECT (transfer_thread),
                   transfer_thread_signals[TARGET_CREATED_SIGNAL],
                   0);

    /* Seek to correct position */
//FIXME: gnome_vfs_async_seek
//FIXME: http://bugzilla.gnome.org/show_bug.cgi?id=123414
//    gnome_vfs_seek (handle, GNOME_VFS_SEEK_START,
//                    transfer_thread->bytes_start);

//We have this in gnome-vfs-method.h:
//#define VFS_METHOD_HAS_FUNC(method,func) ((((char *)&((method)->func)) - ((char *)(method)) < (method)->method_table_size) && method->func != NULL)
//so an async_seek returning GNOME_VFS_ERROR_NOT_SUPPORTED will inform us about
//resume support.

    gnome_vfs_async_open_uri (&transfer_thread->source_handle,
                              transfer_thread->source_uri,
                              GNOME_VFS_OPEN_READ,
//FIXME: Detect if the protocol can OPEN_RANDOM / seek
//FIXME: FIX gnome-vfs on ftp (http already has a patch)!!!
//                              | GNOME_VFS_OPEN_RANDOM,
                              GNOME_VFS_PRIORITY_DEFAULT,
                              transfer_thread_open_source_uri,
                              transfer_thread);
}

void
transfer_thread_open_source_uri (GnomeVFSAsyncHandle *handle,
                                 GnomeVFSResult result,
                                 gpointer data)
{
    TransferThread *transfer_thread = TRANSFER_THREAD (data);

    if (result != GNOME_VFS_OK)
    {
        transfer_thread->status = ERROR_OPENING_SOURCE;
        g_signal_emit (G_OBJECT (transfer_thread),
                       transfer_thread_signals[ERROR_SIGNAL],
                       0);
        return;
    }

    /* Seek to correct position */
//FIXME: gnome_vfs_async_seek
//FIXME: http://bugzilla.gnome.org/show_bug.cgi?id=123414
//    gnome_vfs_seek (handle, GNOME_VFS_SEEK_START,
//                    transfer_thread->bytes_start);

    transfer_thread->status = STATUS_TRANSFERING;
    g_signal_emit (G_OBJECT (transfer_thread),
                   transfer_thread_signals[TRANSFERING_SIGNAL],
                   0);

    gnome_vfs_async_read (transfer_thread->source_handle,
                          transfer_thread->buffer,
                          4096,
                          transfer_thread_source_read,
                          transfer_thread);
}

void
transfer_thread_source_read (GnomeVFSAsyncHandle *handle,
                             GnomeVFSResult result,
                             gpointer buffer,
                             GnomeVFSFileSize bytes_requested,
                             GnomeVFSFileSize bytes_read,
                             gpointer data)
{
    TransferThread *transfer_thread = TRANSFER_THREAD (data);

    if (result == GNOME_VFS_ERROR_EOF)
    {
        transfer_thread->stop = TRUE;
        if ((transfer_thread->bytes_done + bytes_read) !=
            (transfer_thread->bytes_stop - transfer_thread->bytes_start))
        {
            transfer_thread->status = ERROR_READING_SOURCE;
        }
    }
    else if (result != GNOME_VFS_OK)
    {
        transfer_thread->status = ERROR_READING_SOURCE;
        g_signal_emit (G_OBJECT (transfer_thread),
                       transfer_thread_signals[ERROR_SIGNAL],
                       0);
        return;
    }

    gnome_vfs_async_write (transfer_thread->target_handle,
                           buffer,
                           bytes_read,
                           transfer_thread_target_write,
                           transfer_thread);
}

void
transfer_thread_target_write (GnomeVFSAsyncHandle *handle,
                              GnomeVFSResult result,
                              gconstpointer buffer,
                              GnomeVFSFileSize bytes_requested,
                              GnomeVFSFileSize bytes_written,
                              gpointer data)
{
    TransferThread *transfer_thread = TRANSFER_THREAD (data);

    if (result != GNOME_VFS_OK)
    {
        transfer_thread->status = ERROR_WRITING_TARGET;
        g_signal_emit (G_OBJECT (transfer_thread),
                       transfer_thread_signals[ERROR_SIGNAL],
                       0);
        return;
    }

//FIXME: check for bytes_requested != bytes_written
    transfer_thread->bytes_done += bytes_written;

    if (transfer_thread->stop)
    {
//FIXME: Free resources
        if (transfer_thread->bytes_done ==
            (transfer_thread->bytes_stop - transfer_thread->bytes_start))
        {
            transfer_thread->status = STATUS_DONE;
            g_signal_emit (G_OBJECT (transfer_thread),
                           transfer_thread_signals[DONE_SIGNAL],
                           0);
        }
        else
        {
            if (transfer_thread->status == STATUS_TRANSFERING)
                transfer_thread->status = STATUS_STOPPING;
            g_signal_emit (G_OBJECT (transfer_thread),
                           transfer_thread_signals[STOP_SIGNAL],
                           0);
        }
        return;
    }

    if (transfer_thread->speedlimit)
    {
        transfer_thread->speedlimit_left -= bytes_written;

        /* if there is no speedlimit_left, we must wait till the next second */
        if (transfer_thread->speedlimit_left <= 0)
        {
            struct timeval time;
            gettimeofday (&time, NULL);

            /* add a new slice of data to consume */
            transfer_thread->speedlimit_left += transfer_thread->speedlimit;

            /* plus another slice if we missed two seconds */
            if (time.tv_sec + 1 < transfer_thread->oldtime.tv_sec)
                transfer_thread->speedlimit_left += transfer_thread->speedlimit;

            g_timeout_add (1000 - (time.tv_usec / 1000),
                           transfer_thread_restart_transfering,
                           transfer_thread);
        }
        else
        {
            gnome_vfs_async_read (transfer_thread->source_handle,
                                  transfer_thread->buffer,
                                  1024,
                                  transfer_thread_source_read,
                                  transfer_thread);
        }
    }
    else
    {
        /* No speedlimit, do normal reads */
        gnome_vfs_async_read (transfer_thread->source_handle,
                              transfer_thread->buffer,
                              4096,
                              transfer_thread_source_read,
                              transfer_thread);
    }
}

gboolean
transfer_thread_restart_transfering (gpointer data)
{
    TransferThread *transfer_thread = TRANSFER_THREAD (data);

    gnome_vfs_async_read (transfer_thread->source_handle,
                          transfer_thread->buffer,
//FIXME: probably adapt this dinamically (smb most)
                          1024,
                          transfer_thread_source_read,
                          transfer_thread);

    return FALSE;
}

void
transfer_thread_target_uri_set (TransferThread *transfer_thread,
                                GnomeVFSURI *target_uri)
{
    transfer_thread->target_uri = target_uri;
}

void
transfer_thread_source_uri_set (TransferThread *transfer_thread,
                                GnomeVFSURI *source_uri)
{
    transfer_thread->source_uri = source_uri;
}

void
transfer_thread_bytes_start_set (TransferThread *transfer_thread,
                                 GnomeVFSFileSize bytes_start)
{
    transfer_thread->bytes_start = bytes_start;
}

void
transfer_thread_bytes_stop_set (TransferThread *transfer_thread,
                                GnomeVFSFileSize bytes_stop)
{
    transfer_thread->bytes_stop = bytes_stop;
}

GnomeVFSFileSize
transfer_thread_bytes_done_get (TransferThread *transfer_thread)
{
    return transfer_thread->bytes_done;
}

void
transfer_thread_speedlimit_set (TransferThread *transfer_thread,
                                guint speedlimit)
{
    transfer_thread->speedlimit = speedlimit;
    transfer_thread->speedlimit_left = speedlimit;
}

void
transfer_thread_target_create (TransferThread *transfer_thread,
                               gboolean create_target)
{
    transfer_thread->create_target = create_target;
}
