#include <sys/time.h>

#include <glib.h>
#include <libgnomevfs/gnome-vfs-ops.h>
#include <libgnomevfs/gnome-vfs-uri.h>

#include <utillib.h>

#include "debug.h"
#include "transfer.h"

#include "downmand-marshal.h"

#include "transfer-thread.h"

GType           transfer_get_type           ();
static void     transfer_class_init         (TransferClass *klass);
static void     transfer_init               (Transfer *transfer);
static void     transfer_finalize           (GObject* object);

void            transfer_thread_exit        (Transfer *transfer,
                                             TransferThread *thread);

enum {
    TRANSFER_UPDATE,
    TRANSFER_EXIT,
    LAST_SIGNAL
};

static gint transfer_signals[LAST_SIGNAL] = { 0 };

static gpointer parent_class = NULL;


gpointer transfer_thread(gpointer data);

void    transfer_cleanup        (void *arg);
gint    resolve                 (Transfer *transfer, TransferThread *thread);
gint    create_socket           (Transfer *transfer, TransferThread *thread);
gint    connect_to_host         (Transfer *transfer, TransferThread *thread);

void    transfer_get_target_info_callback   (GnomeVFSAsyncHandle *handle,
                                             GList *results,
                                             gpointer data);
void    transfer_get_source_info_callback   (GnomeVFSAsyncHandle *handle,
                                             GList *results,
                                             gpointer data);

void    transfer_target_created             (TransferThread *transfer_thread,
                                             gpointer data);
void    transfer_thread_transfering         (TransferThread *transfer_thread,
                                             gpointer data);
void    transfer_thread_stopped             (TransferThread *transfer_thread,
                                             gpointer data);
void    transfer_thread_finished            (TransferThread *transfer_thread,
                                             gpointer data);
void    transfer_thread_error               (TransferThread *transfer_thread,
                                             gpointer data);
gboolean    transfer_retry_again            (gpointer data);


GType transfer_get_type ()
{
    static GType transfer_type = 0;

    if (!transfer_type)
    {
        static const GTypeInfo transfer_info =
        {
            sizeof (TransferClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) transfer_class_init,
            NULL,           /* class_finalize */
            NULL,           /* class_data */
            sizeof (Transfer),
            0,              /* n_preallocs */
            (GInstanceInitFunc) transfer_init
        };

        transfer_type = g_type_register_static (G_TYPE_OBJECT,
                                               "Transfer",
                                               &transfer_info,
                                               0);
    }

    return transfer_type;
}

static void transfer_class_init (TransferClass *klass)
{
    GObjectClass *object_class;

    object_class = (GObjectClass*) klass;

    parent_class = g_type_class_peek_parent (klass);

    transfer_signals[TRANSFER_UPDATE] = g_signal_new ("transfer_update",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (TransferClass, transfer_update_signal),
            NULL, NULL,
            downmand_marshal_VOID__INT,
            G_TYPE_NONE,
            1, G_TYPE_INT);

    transfer_signals[TRANSFER_EXIT] = g_signal_new ("transfer_exit",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (TransferClass, transfer_exit_signal),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    klass->transfer_update_signal = NULL;
    klass->transfer_exit_signal = NULL;

    object_class->finalize  = transfer_finalize;
}

static void
transfer_init (Transfer *transfer)
{
    transfer->id = 0;
    transfer->order = 0;

    transfer->threads = NULL;

    transfer->bytes_done = 0;
    transfer->bytes_total = 0;
    transfer->speed = 0;
    transfer->speedlimit = 0;
    transfer->real_speedlimit = 0;
    transfer->target_ready = 0;
    transfer->queue = TRANSFER_QUEUE_NONE;
    transfer->status = STATUS_QUEUED;
    transfer->exit_cause = TRANSFER_EXIT_NONE;

    transfer->target_uri = 0;
    transfer->source_uri = 0;
    transfer->target_info = 0;
    transfer->source_info = 0;
}

static void
transfer_finalize (GObject* object)
{
    Transfer *transfer = TRANSFER (object);

    gnome_vfs_uri_unref (transfer->source_uri);
    gnome_vfs_uri_unref (transfer->target_uri);

    if (G_OBJECT_CLASS (parent_class)->finalize)
        (* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

Transfer *transfer_new ()
{
    Transfer *transfer = TRANSFER (g_object_new (transfer_get_type (), NULL));

    transfer->linelog = linelog_new ();

    return transfer;
}

gboolean
transfer_source_uri_set (Transfer *transfer, gchar *text_uri)
{
    GnomeVFSURI *uri;

    uri = gnome_vfs_uri_new (text_uri);

    if (!uri)
        return FALSE;

    if (transfer->source_uri)
        gnome_vfs_uri_unref (transfer->source_uri);

    transfer->source_uri = uri;

    return TRUE;
}

gboolean
transfer_target_uri_set (Transfer *transfer, gchar *text_uri)
{
    GnomeVFSURI *uri;

    uri = gnome_vfs_uri_new (text_uri);

    if (!uri)
        return FALSE;

    if (transfer->target_uri)
        gnome_vfs_uri_unref (transfer->target_uri);

    transfer->target_uri = uri;

    return TRUE;
}

gboolean
transfer_path_set (Transfer *transfer, gchar *text_path)
{
    GnomeVFSURI *uri;

    uri = gnome_vfs_uri_new (text_path);

    if (!uri)
        return FALSE;

    if (transfer->target_path)
        gnome_vfs_uri_unref (transfer->target_path);

    transfer->target_path = uri;

    return TRUE;
}

void
transfer_start (Transfer *transfer, GError **error)
{
    GList *uri_list = NULL;

    if (!transfer->source_uri)
    {
        g_set_error (error, DM_DOWNMAND_ERROR, 1,
            "The source URI is invalid");
        return;
    }

    if (!transfer->target_uri)
    {
        if ((transfer->source_uri->text[strlen (transfer->source_uri->text)] == '/'))
        {
            transfer->target_uri = gnome_vfs_uri_append_file_name (
                                                        transfer->target_path,
                                                        "index.html");
        }
        else
        {
            transfer->target_uri = gnome_vfs_uri_append_file_name (
                    transfer->target_path,
                    gnome_vfs_uri_extract_short_name (transfer->source_uri));
        }
    }

    transfer->status = STATUS_INIT;
    transfer->exit_cause = TRANSFER_EXIT_NONE;

    /* Check if the target uri exists */
    uri_list = g_list_append (uri_list, transfer->target_uri);
    gnome_vfs_async_get_file_info (&transfer->async_handle,
                                   uri_list,
                                   GNOME_VFS_FILE_INFO_DEFAULT,
                                   GNOME_VFS_PRIORITY_DEFAULT,
                                   transfer_get_target_info_callback,
                                   transfer);
}

void
transfer_get_target_info_callback (GnomeVFSAsyncHandle *handle, GList *results,
                                   gpointer data)
{
    Transfer *transfer = TRANSFER (data);
    GnomeVFSGetFileInfoResult* info_result;
    GList *uri_list = NULL;

    if (results == NULL)
        return;

    if (transfer->target_info)
        g_free (transfer->target_info);

    info_result = results->data;

    transfer->target_info = gnome_vfs_file_info_dup (info_result->file_info);

    if (info_result->result == GNOME_VFS_OK)
    {
        if (!transfer->target_created)
        {
            /* We don't created the file, ask the user */
//FIXME: ask the user what to do: overwrite (y/n)?
            transfer->status = ERROR_OPENING_TARGET;
        }
    }
    else if (info_result->result == GNOME_VFS_ERROR_NOT_FOUND)
    {
        /* The file does not exist, do we created it? */
        if (transfer->target_created)
        {
            /* We created the file, something is wrong, ask the user */
//FIXME: Tell the user what to do: restart transfer?
            transfer->status = ERROR_OPENING_TARGET;
        }
    }
    else
    {
        transfer->status = ERROR_OPENING_TARGET;
    }

    if (transfer->status == ERROR_OPENING_TARGET)
    {
        /* Stop this thread */
//FIXME: free resources
        return;
    }

    /* Check if the source uri exists */
    uri_list = g_list_append (uri_list, transfer->source_uri);
    gnome_vfs_async_get_file_info (&transfer->async_handle,
                                   uri_list,
                                   GNOME_VFS_FILE_INFO_DEFAULT,
                                   GNOME_VFS_PRIORITY_DEFAULT,
                                   transfer_get_source_info_callback,
                                   transfer);
}

void
transfer_get_source_info_callback (GnomeVFSAsyncHandle *handle, GList *results,
                                   gpointer data)
{
    Transfer *transfer = TRANSFER (data);
    GnomeVFSGetFileInfoResult* info_result;
    TransferThread *transfer_thread;

    if (results == NULL)
        return;

    if (transfer->target_info)
        g_free (transfer->target_info);

    info_result = results->data;

    transfer->source_info = gnome_vfs_file_info_dup (info_result->file_info);

    if (info_result->result == GNOME_VFS_OK)
    {
        transfer->status = STATUS_STARTING;
    }
    else if (info_result->result == GNOME_VFS_ERROR_NOT_FOUND)
    {
        transfer->status = ERROR_OPENING_SOURCE;
    }
    else
    {
        transfer->status = ERROR_OPENING_SOURCE;
    }

//FIXME: user interaction on errors? perhaps just stop.
    if (transfer->status == ERROR_OPENING_SOURCE)
    {
        /* Stop this thread */
//FIXME: free resources.
        return;
    }

//FIXME: we need async_seek support
	transfer->bytes_done = 0;

//FIXME: this must be changed for multi-thread transfers
    /* Check if we are already done */
    if (transfer->source_info->size > 0)
    {
        if (transfer->target_info->size == transfer->source_info->size)
        {
            transfer->exit_cause = TRANSFER_EXIT_DONE;
//FIXME: Free resources and generate signal
        }
        if (transfer->target_info->size > transfer->source_info->size)
        {
            transfer->exit_cause = TRANSFER_EXIT_STOPPED;
//FIXME: Free resources and generate signal
        }
    }

    /* If the target size is > 0, we need seek() support */
//FIXME: gnome-vfs: method abilities
//http://bugzilla.gnome.org/show_bug.cgi?id=123821
/*
    if (transfer->target_info->size > 0)
    {
        struct GnomeVFSMethod *method;

        method = gnome_vfs_method_get (transfer->source_uri->method_string);
        if (method)
        {
            if (method->GnomeVFSMethodSeekFunc == NULL)
            {
                ...
                return;
            }
        }
    }
*/

    transfer->bytes_total = transfer->source_info->size;
    transfer->oldtime.tv_sec = 0;

    transfer_thread = transfer_thread_new ();

    transfer_thread_target_uri_set (transfer_thread, transfer->target_uri);
    transfer_thread_source_uri_set (transfer_thread, transfer->source_uri);
    transfer_thread_bytes_start_set (transfer_thread, 0);
    transfer_thread_bytes_stop_set (transfer_thread,
                                    transfer->source_info->size);
    transfer_thread_target_create (transfer_thread, !transfer->target_created);
//FIXME: multi-thread transfers need some speedlimit division
    transfer_thread_speedlimit_set (transfer_thread, transfer->real_speedlimit);

    transfer->threads = g_list_append (transfer->threads, transfer_thread);

    g_signal_connect (transfer_thread, "target_created_signal",
                      G_CALLBACK (transfer_target_created),
                      transfer);
    g_signal_connect (transfer_thread, "transfering_signal",
                      G_CALLBACK (transfer_thread_transfering),
                      transfer);
    g_signal_connect (transfer_thread, "stop_signal",
                      G_CALLBACK (transfer_thread_stopped),
                      transfer);
    g_signal_connect (transfer_thread, "done_signal",
                      G_CALLBACK (transfer_thread_finished),
                      transfer);
    g_signal_connect (transfer_thread, "error_signal",
                      G_CALLBACK (transfer_thread_error),
                      transfer);

    transfer_thread_start (transfer_thread);
}

void
transfer_target_created (TransferThread *transfer_thread, gpointer data)
{
    Transfer *transfer = TRANSFER (data);

    transfer->target_created = TRUE;
}

void
transfer_thread_transfering (TransferThread *transfer_thread, gpointer data)
{
    Transfer *transfer = TRANSFER (data);

    transfer->status = STATUS_TRANSFERING;
}

void
transfer_thread_stopped (TransferThread *transfer_thread, gpointer data)
{
    Transfer *transfer = TRANSFER (data);

    transfer_statistics (transfer);
    transfer->threads = g_list_remove (transfer->threads,
                                       transfer->threads->data);

    if (transfer_thread->status == STATUS_STOPPING)
        transfer->status = STATUS_STOPPED;
    else
        transfer->status = transfer_thread->status;

    transfer->exit_cause = TRANSFER_EXIT_STOPPED;

    g_signal_emit (G_OBJECT (transfer),
                   transfer_signals[TRANSFER_EXIT],
                   0);
}

void
transfer_thread_finished (TransferThread *transfer_thread, gpointer data)
{
    Transfer *transfer = TRANSFER (data);

    transfer_statistics (transfer);

    transfer->status = STATUS_DONE;
    transfer->exit_cause = TRANSFER_EXIT_DONE;
    transfer->speed = 0;
    g_signal_emit (G_OBJECT (transfer),
                   transfer_signals[TRANSFER_EXIT],
                   0);
}

void
transfer_thread_error (TransferThread *transfer_thread, gpointer data)
{
    Transfer *transfer = TRANSFER (data);

    transfer_statistics (transfer);

    transfer->status = transfer_thread->status;
    transfer->speed = 0;
    switch (transfer_thread->status)
    {
        case ERROR_OPENING_SOURCE:
        case ERROR_OPENING_TARGET:
//FIXME: user interaction
            transfer->exit_cause = TRANSFER_EXIT_STOPPED;
            g_signal_emit (G_OBJECT (transfer),
                           transfer_signals[TRANSFER_EXIT],
                           0);
            break;

        default:
            /* Wait some seconds and try again */
//FIXME: Max. retries
            g_timeout_add (10 * 1000, transfer_retry_again, transfer);
            break;
    }
}

gboolean
transfer_retry_again (gpointer data)
{
    Transfer *transfer = TRANSFER (data);
    GError *err = NULL;

    transfer_start (transfer, &err);

    return FALSE;
}

void
transfer_statistics (Transfer *transfer)
{
    TransferThread *transfer_thread;
	GnomeVFSFileSize bytes_done_previous;
	struct timeval newtime;
    gint seconds_elapsed;

    if (!transfer->threads)
        return;

    if (!transfer->threads->data)
        return;

    transfer_thread = transfer->threads->data;
    bytes_done_previous = transfer->bytes_done;
    transfer->bytes_done = transfer_thread_bytes_done_get (transfer_thread);

    /* speed statistics */
    gettimeofday (&newtime, NULL);
    seconds_elapsed = newtime.tv_sec - transfer->oldtime.tv_sec;
    if (!seconds_elapsed)
        return;

    transfer->oldtime = newtime;

    if (seconds_elapsed < 6)
    {
        guint speed;
        gint i;

        transfer->speeds[newtime.tv_sec % 5] =
                (transfer->bytes_done - bytes_done_previous) / seconds_elapsed;
        speed = 0;
        for (i = 0; i < 5; i++)
        {
            speed += transfer->speeds[i];
        }
        speed /= 5;
        transfer->speed = speed;
    }
    else
    {
        transfer->speed = transfer->bytes_done - bytes_done_previous;
        transfer->speeds[0] = transfer->speed;
        transfer->speeds[1] = transfer->speed;
        transfer->speeds[2] = transfer->speed;
        transfer->speeds[3] = transfer->speed;
        transfer->speeds[4] = transfer->speed;
    }
}

void transfer_thread_exit(Transfer *transfer, TransferThread *thread)
{
    GError *error;

//FIXME: per thread exit_cause
    switch (transfer->exit_cause)
    {
        case TRANSFER_EXIT_STOPPED:
            transfer->status = STATUS_STOPPED;
            break;
        case TRANSFER_EXIT_QUEUED:
            transfer->status = STATUS_QUEUED;
            break;
        case TRANSFER_EXIT_DONE:
            transfer->status = STATUS_DONE;
            break;
        case TRANSFER_EXIT_SHUTDOWN:
            transfer->status = STATUS_STOPPED;
            break;

        default:
            break;
    }

    g_signal_emit (G_OBJECT (transfer),
               transfer_signals[TRANSFER_EXIT],
               0);

    linelog_disklog (transfer->linelog, FALSE, &error);

    g_thread_exit(NULL);
}


void
transfer_stop (Transfer *transfer, TransferStopAction action)
{
    TransferThread *transfer_thread;

    switch (transfer->queue)
    {
        case TRANSFER_QUEUE_ACTIVE:
            switch(action)
            {
                case TRANSFER_STOP_NONE:
                    transfer->exit_cause = TRANSFER_EXIT_STOPPED;
                    break;
                case TRANSFER_STOP_QUEUE:
                    transfer->exit_cause = TRANSFER_EXIT_QUEUED;
                    break;
                case TRANSFER_STOP_DELETE:
                    transfer->exit_cause = TRANSFER_EXIT_DELETE;
                    break;
                case TRANSFER_STOP_SHUTDOWN:
                    transfer->exit_cause = TRANSFER_EXIT_SHUTDOWN;
                    break;
                default:
                    break;
            }
            transfer->status = STATUS_STOPPING;

//FIXME: this must be changed for multiple threads
            if (!transfer->threads)
                return;

            if (!transfer->threads->data)
                return;

            transfer_thread = transfer->threads->data;

//FIXME: use method
            transfer_thread->stop = TRUE;
            break;

        default:
            break;
    }
}

void
transfer_logdir_set (Transfer *transfer, gchar *path)
{
    GString *buffer = g_string_new (NULL);

	g_string_printf (buffer, "%s/%u.txt", path, transfer->id);
    linelog_setfile (transfer->linelog, buffer->str);

    g_string_free(buffer, TRUE);
}

void
transfer_log_add (Transfer *transfer, const gchar *format, ...)
{
    GString *logger = g_string_new (NULL);
    GError *error = NULL;
    va_list args;

    va_start (args, format);
    g_string_printf (logger, format, args);
    va_end (args);

    linelog_add (transfer->linelog, logger, &error);
//FIXME: log the error
}

gint transfer_status_get (Transfer *transfer)
{
    return transfer->status;
}

void transfer_status_set(Transfer *transfer, enum dm_transfer_status status)
{
	transfer->status = status;
}

guint
transfer_recursive_get (Transfer *transfer)
{
    return transfer->recursive;
}

void
transfer_recursive_set (Transfer *transfer, guint recursive)
{
    transfer->recursive = recursive;
}

void
transfer_default_path_set (Transfer *transfer, gboolean setting)
{
    transfer->default_path = setting;
}

void
transfer_speedlimit_set (Transfer *transfer, guint speed)
{
    transfer->speedlimit = speed;

    /* We send the command that is used to modify the property */
    g_signal_emit (G_OBJECT (transfer), transfer_signals [TRANSFER_UPDATE], 0,
                   DM_COMMAND_ITEM_SPEED);
}

guint
transfer_speedlimit_get (Transfer *transfer)
{
    return transfer->speedlimit;
}

void
transfer_real_speedlimit_set (Transfer *transfer, guint speed)
{
    transfer->real_speedlimit = speed;

    if (!transfer->threads)
        return;

    if (!transfer->threads->data)
        return;

//FIXME: multi-thread transfers need some speedlimit division
    transfer_thread_speedlimit_set (transfer->threads->data, speed);
}

guint
transfer_real_speedlimit_get (Transfer *transfer)
{
    return transfer->real_speedlimit;
}

TransferExitCause transfer_exit_cause(Transfer *transfer)
{
	return transfer->exit_cause;
}

void
transfer_id_set (Transfer *transfer, guint id)
{
    transfer->id = id;
}

guint
transfer_id_get (Transfer *transfer)
{
    return transfer->id;
}

gchar *
transfer_source_uri_string (Transfer *transfer)
{
    gchar *uri = NULL;

    if (transfer->source_uri)
        uri = gnome_vfs_uri_to_string (transfer->source_uri,
                                       GNOME_VFS_URI_HIDE_PASSWORD);

    return uri;
}

gchar *
transfer_target_uri_string (Transfer *transfer)
{
    gchar *uri = NULL;

    if (transfer->target_uri)
        uri = gnome_vfs_uri_to_string (transfer->target_uri,
                                       GNOME_VFS_URI_HIDE_PASSWORD);

    return uri;
}

gchar *
transfer_targetpath (Transfer *transfer)
{
    gchar *uri = NULL;

    if (transfer->target_path)
        uri = gnome_vfs_uri_to_string (transfer->target_path,
                                       GNOME_VFS_URI_HIDE_PASSWORD);

    return uri;
}

GnomeVFSFileSize
transfer_bytes_done_get (Transfer *transfer)
{
    return transfer->bytes_done;
}

void
transfer_bytes_done_set (Transfer *transfer, GnomeVFSFileSize bytes_done)
{
    transfer->bytes_done = bytes_done;
}

GnomeVFSFileSize
transfer_bytes_total_get (Transfer *transfer)
{
    return transfer->bytes_total;
}

void
transfer_bytes_total_set (Transfer *transfer, GnomeVFSFileSize bytes_total)
{
    transfer->bytes_total = bytes_total;
}

guint
transfer_speed (Transfer *transfer)
{
    return transfer->speed;
}

TransferQueue transfer_queue_get (Transfer *transfer)
{
	return transfer->queue;
}

void transfer_queue_set (Transfer *transfer, TransferQueue queue)
{
	transfer->queue = queue;
}

void
transfer_target_created_set (Transfer *transfer, gboolean created)
{
    transfer->target_created = created;
}

gboolean
transfer_target_created_get (Transfer *transfer)
{
    return transfer->target_created;
}
